@@ -39,24 +39,13 @@ else:
 
 
 # in case of changes in the way cachew stores data, this should be changed to discard old caches
-CACHEW_FORMAT = 1
+CACHEW_VERSION: str = __version__
 
 
 def get_logger() -> logging.Logger:
     return logging.getLogger('cachew')
 
 
-T = TypeVar('T')
-
-
-def ichunks(l: Iterable[T], n: int) -> Iterator[List[T]]:
-    it: Iterator[T] = iter(l)
-    while True:
-        chunk: List[T] = list(islice(it, 0, n))
-        if len(chunk) == 0:
-            break
-        yield chunk
-
 
 class IsoDateTime(sqlalchemy.TypeDecorator):
     # in theory could use something more effecient? e.g. blob for encoded datetime and tz?
@@ -462,7 +451,29 @@ SourceHash = str
 
 class DbHelper:
     def __init__(self, db_path: Path, cls: Type) -> None:
-        self.engine = sqlalchemy.create_engine(f'sqlite:///{db_path}')
+        self.engine = sqlalchemy.create_engine(f'sqlite:///{db_path}', connect_args={'timeout': 0})
+        # NOTE: timeout is necessary so we don't lose time waiting during recursive calls
+        # by default, it's several seconds? you'd see 'test_recursive' test performance degrade
+
+        import sqlite3
+        import time
+
+        from sqlalchemy.engine import Engine # type: ignore
+        @event.listens_for(Engine, 'connect')
+        def set_sqlite_pragma(dbapi_connection, connection_record):
+            # without wal, concurrent reading/writing is not gonna work
+
+            # ugh. that's odd, how are we supposed to set WAL if the very fact of setting wal might lock the db?
+            while True:
+                try:
+                    dbapi_connection.execute('PRAGMA journal_mode=WAL')
+                    break
+                except sqlite3.OperationalError as oe:
+                    if 'database is locked' not in str(oe):
+                        raise oe
+                time.sleep(0.1)
+
+
         self.connection = self.engine.connect()
 
         """
@@ -483,7 +494,6 @@ class DbHelper:
 
         self.meta = sqlalchemy.MetaData(self.connection)
         self.table_hash = Table('hash', self.meta, Column('value', sqlalchemy.String))
-        self.table_hash.create(self.connection, checkfirst=True)
 
         self.binder = NTBinder.make(tp=cls)
         self.table_data = Table('table', self.meta, *self.binder.columns)
@@ -591,10 +601,17 @@ def cachew(
         func=None,
         cache_path: Optional[PathProvider]=None,
         cls=None,
+        # TODO name 'dependencies'? or 'depends_on'?
         hashf: HashFunction=default_hash,
         logger=None,
-        chunk_by=10000, # TODO dunno maybe remove it?
-):  # TODO what's a reasonable default?):
+        chunk_by=100,
+        # NOTE: allowed values for chunk_by depend on the system.
+        # some systems (to be more specific, sqlite builds), it might be too large and cause issues
+        # ideally this would be more defensive/autodetected, maybe with a warning?
+        # you can use 'test_many' to experiment
+        # - too small values (e.g. 10)  are slower than 100 (presumably, too many sql statements)
+        # - too large values (e.g. 10K) are slightly slower as well (not sure why?)
+):
     r"""
     Database-backed cache decorator. TODO more description?
     # TODO use this doc in readme?
@@ -698,7 +715,7 @@ def cachew_impl(*, func: Callable, cache_path: PathProvider, cls: Type, hashf: H
         kwargs = {**defaults, **kwargs}
         # TODO not sure if passing them makes sense??
         # TODO FIXME use inspect.signature to inspect return type annotations at least?
-        return f'cachew: {CACHEW_FORMAT}, schema: {get_schema(cls)}, hash: {hashf(*args, **kwargs)}'
+        return f'cachew: {CACHEW_VERSION}, schema: {get_schema(cls)}, hash: {hashf(*args, **kwargs)}'
 
     @functools.wraps(func)
     def wrapper(*args, **kwargs):
@@ -713,67 +730,97 @@ def cachew_impl(*, func: Callable, cache_path: PathProvider, cls: Type, hashf: H
         if not dbp.parent.exists():
             raise CachewException(f"{dbp.parent} doesn't exist") # otherwise, sqlite error is quite cryptic
 
+        h = composite_hash(*args, **kwargs); kassert(h is not None)  # just in case
+        logger.debug('new hash: %s', h)
 
-        with DbHelper(dbp, cls) as db:
-            binder = db.binder
+        with DbHelper(dbp, cls) as db, \
+             db.connection.begin():
+            # NOTE: defferred transaction
             conn = db.connection
+            binder = db.binder
             values_table = db.table_data
 
-            h = composite_hash(*args, **kwargs); kassert(h is not None)  # just in case
-            logger.debug('new hash: %s', h)
-
-            # transaction is DEFERRED  (see DbHelper constructor)
-            with conn.begin():
+            # first, try to do as much as possible read-only, benefiting from deferred transaction
+            try:
+                # not sure if there is a better way...
                 prev_hashes = conn.execute(db.table_hash.select()).fetchall()
-                kassert(len(prev_hashes) <= 1)  # shouldn't happen
-
-                prev_hash: Optional[SourceHash]
-                if len(prev_hashes) == 0:
-                    prev_hash = None
+            except sqlalchemy.exc.OperationalError as e:
+                # meh. not sure if this is a good way to handle this..
+                if 'no such table: hash' in str(e):
+                    prev_hashes = []
                 else:
-                    prev_hash = prev_hashes[0][0]  # returns a tuple...
+                    raise e
 
-                logger.debug('old hash: %s', prev_hash)
+            kassert(len(prev_hashes) <= 1)  # shouldn't happen
+            prev_hash: Optional[SourceHash]
+            if len(prev_hashes) == 0:
+                prev_hash = None
+            else:
+                prev_hash = prev_hashes[0][0]  # returns a tuple...
 
-                if h == prev_hash:
-                    logger.debug('hash matched: loading from cache')
-                    rows = conn.execute(values_table.select())
-                    for row in rows:
-                        yield binder.from_row(row)
+            logger.debug('old hash: %s', prev_hash)
+
+            if h == prev_hash:
+                logger.debug('hash matched: loading from cache')
+                rows = conn.execute(values_table.select())
+                for row in rows:
+                    yield binder.from_row(row)
+                return
+
+            logger.debug('hash mismatch: computing data and writing to db')
+
+            # NOTE on recursive calls
+            # somewhat magically, they should work as expected with no extra database inserts?
+            # the top level call 'wins' the write transaction and once it's gathered all data, will write it
+            # the 'intermediate' level calls fail to get it and will pass data through
+            # the cached 'bottom' level is read only and will be yielded withotu a write transaction
+
+            try:
+                # first write statement will upgrade transaction to write transaction which might fail due to concurrency
+                # see https://www.sqlite.org/lang_transaction.html
+                # note I guess, because of 'checkfirst', only the last create is actually guaranteed to upgrade the transaction to write one
+                # drop and create to incorporate schema changes
+                db.table_hash.create(conn, checkfirst=True)
+                values_table.drop(conn, checkfirst=True)
+                values_table.create(conn)
+            except sqlalchemy.exc.OperationalError as e:
+                if e.code == 'e3q8':
+                    # database is locked; someone else must be have won the write lock
+                    # not much we can do here
+                    # NOTE: important to close early, otherwise we might hold onto too many file descriptors during yielding
+                    # see test_deep_recursive
+                    conn.close()
+                    yield from func(*args, **kwargs)
+                    return
                 else:
-                    logger.debug('hash mismatch: computing data and writing to db')
+                    raise e
+            # at this point we're guaranteed to have an exclusive write transaction
 
-                    # first write statement will upgrade transaction to write transaction which might fail due to concurrency
-                    # see https://www.sqlite.org/lang_transaction.html
-                    # TODO i'd prefer some explicit transaction upgrade statement; but not sure if there is any?
-                    try:
-                        # drop and create to incorporate schema changes
-                        values_table.drop(conn, checkfirst=True)
-                    except sqlalchemy.exc.OperationalError as e:
-                        if e.code == 'e3q8':
-                            # database is locked; someone else must be writing
-                            # not much we can do here
-                            yield from func(*args, **kwargs)
-                            return
-                        else:
-                            raise e
-                    values_table.create(conn)
-
-                    datas = func(*args, **kwargs)
-
-                    for chunk in ichunks(datas, n=chunk_by):
-                        bound = [binder.to_row(c) for c in chunk]
-                        # pylint: disable=no-value-for-parameter
-                        conn.execute(values_table.insert().values(bound))
-                        yield from chunk
-
-                    # TODO insert and replace instead?
+            datas = func(*args, **kwargs)
 
+            chunk: List[Any] = []
+            def flush():
+                nonlocal chunk
+                if len(chunk) > 0:
                     # pylint: disable=no-value-for-parameter
-                    conn.execute(db.table_hash.delete())
-                    # pylint: disable=no-value-for-parameter
-                    conn.execute(db.table_hash.insert().values([{'value': h}]))
+                    conn.execute(values_table.insert().values(chunk))
+                    chunk = []
+
+            for d in datas:
+                yield d
+                chunk.append(binder.to_row(d))
+                if len(chunk) >= chunk_by:
+                    flush()
+            flush()
+
+            # TODO insert and replace instead?
+
+            # pylint: disable=no-value-for-parameter
+            conn.execute(db.table_hash.delete())
+            # pylint: disable=no-value-for-parameter
+            conn.execute(db.table_hash.insert().values([{'value': h}]))
     return wrapper
 
 
 __all__ = ['cachew', 'CachewException', 'SourceHash', 'HashFunction', 'get_logger', 'NTBinder']
+# TODO add test for migration? actually commit a db in the repository?
