@@ -1,6 +1,8 @@
 from datetime import datetime, date
+import logging
 from pathlib import Path
 from random import Random
+from subprocess import check_call
 import string
 import sys
 import time
@@ -13,6 +15,9 @@ import pytest  # type: ignore
 from .. import cachew, get_logger, PRIMITIVES, NTBinder, CachewException, Types, Values
 
 
+logger = get_logger()
+
+
 @pytest.fixture
 def tdir(tmp_path):
     # TODO just use tmp_path instead?
@@ -548,18 +553,18 @@ def fuzz_cachew_impl():
     import patchy  # type: ignore[import]
     from .. import cachew_impl
     patch = '''\
-@@ -47,6 +47,11 @@
+@@ -740,6 +740,11 @@
 
-                 logger.debug('old hash: %s', prev_hash)
+             logger.debug('old hash: %s', prev_hash)
 
-+                from random import random
-+                rs = random() * 2
-+                print("sleeping for: ", rs)
-+                from time import sleep; sleep(rs)
++            from random import random
++            rs = random() * 2
++            print("sleeping for: ", rs)
++            from time import sleep; sleep(rs)
 +
-                 if h == prev_hash:
-                     logger.debug('hash matched: loading from cache')
-                     rows = conn.execute(values_table.select()
+             if h == prev_hash:
+                 logger.debug('hash matched: loading from cache')
+                 rows = conn.execute(values_table.select())
 '''
     patchy.patch(cachew_impl, patch)
     yield
@@ -631,6 +636,81 @@ def test_mcachew(tmp_path: Path):
     assert list(func()) == ['one', 'two']
 
 
+def test_recursive(tmp_path: Path):
+    calls = 0
+    @cachew(tmp_path)
+    def factorials(n: int) -> Iterable[int]:
+        nonlocal calls
+        calls += 1
+        if n == 0:
+            yield 1
+            return
+        prev = factorials(n - 1)
+        last = 1
+        # TODO potentially quadratic? measure perf perhaps?
+        for x in prev:
+            yield x
+            last = x
+        yield last * n
+
+    assert calls == 0
+    assert list(factorials(3)) == [1, 1, 2, 6]
+    assert calls == 4
+    assert list(factorials(3)) == [1, 1, 2, 6]
+    assert calls == 4
+    assert list(factorials(5)) == [1, 1, 2, 6, 24, 120]
+    assert calls == 6
+    assert list(factorials(3)) == [1, 1, 2, 6]
+    assert calls == 10
+
+
+
+def test_deep_recursive(tmp_path: Path):
+    @cachew(tmp_path)
+    def rec(n: int) -> Iterable[int]:
+        if n == 0:
+            yield 0
+            return
+        yield from rec(n - 1)
+        yield n
+
+    rlimit = sys.getrecursionlimit()
+
+    # NOTE in reality it has to do with the number of file descriptors (ulimit -Sn, e.g. 1024?)
+    # but it seems that during the error unrolling, pytest or something else actually hits the recursion limit somehow
+    # pytest ends up with an internal error in such case... which is good enough as long as tests are concerned I guess.
+    try:
+        sys.setrecursionlimit(2000)
+        list(rec(800))
+        list(rec(800))
+    finally:
+        sys.setrecursionlimit(rlimit)
+
+
+def test_recursive_error(tmp_path: Path):
+    @cachew(tmp_path)
+    def rec(n: int) -> Iterable[int]:
+        if n == 0:
+            yield 0
+            return
+        yield from rec(n - 1)
+        yield n
+
+    rlimit = sys.getrecursionlimit()
+    try:
+        sys.setrecursionlimit(50)
+        list(rec(100))
+        raise AssertionError('Expecting recursion error')
+    except RecursionError:
+        pass
+    finally:
+        sys.setrecursionlimit(rlimit)
+
+    # todo not sure if cache file should exist??
+    # either way, at least check that the db is not completely messed up
+    assert len(list(rec(100))) == 101
+
+
 @pytest.fixture
 def with_exceptions():
     from cachew.experimental import enable_exceptions, disable_exceptions
@@ -667,3 +747,79 @@ def test_result(tmp_path: Path, with_exceptions):
     # looks bit sad indeed, but whatever...
     # TODO need to warn perhaps
 
+
+def test_version_change(tmp_path: Path):
+    calls = 0
+    @cachew(tmp_path, logger=logger)
+    def fun() -> Iterator[str]:
+        nonlocal calls
+        calls += 1
+
+        yield from ['a', 'b', 'c']
+
+    list(fun())
+    list(fun())
+    assert calls == 1
+
+    # todo ugh. not sure how to do this as a relative import??
+    import cachew as cachew_module
+    old_version = cachew_module.CACHEW_VERSION
+
+    try:
+        cachew_module.CACHEW_VERSION = old_version + '_whatever'
+        # should invalidate cachew now
+        list(fun())
+        assert calls == 2
+        list(fun())
+        assert calls == 2
+    finally:
+        cachew_module.CACHEW_VERSION = old_version
+
+    # and now again, back to the old version
+    list(fun())
+    assert calls == 3
+    list(fun())
+    assert calls == 3
+
+
+def dump_old_cache(tmp_path: Path):
+    # call this if you want to get an sql script for version upgrade tests..
+    oc = tmp_path / 'old_cache.sqlite'
+    @cachew(oc)
+    def fun() -> Iterator[int]:
+        yield from [1, 2, 3]
+
+    list(fun())
+    assert oc.exists(), oc
+    from subprocess import check_output, check_call
+    sql = check_output(['sqlite3', oc, '.dump']).decode('utf8')
+    print(sql, file=sys.stderr)
+
+
+
+def test_old_cache_v0_6_3(tmp_path: Path):
+    sql = '''
+PRAGMA foreign_keys=OFF;
+BEGIN TRANSACTION;
+CREATE TABLE hash (
+	value VARCHAR
+);
+INSERT INTO hash VALUES('cachew: 1, schema: {''_'': <class ''int''>}, hash: ()');
+CREATE TABLE IF NOT EXISTS "table" (
+	_cachew_primitive INTEGER
+);
+INSERT INTO "table" VALUES(1);
+INSERT INTO "table" VALUES(2);
+INSERT INTO "table" VALUES(3);
+COMMIT;
+    '''
+    db = tmp_path / 'cache.sqlite'
+    check_call(['sqlite3', db, sql])
+
+    @cachew(db)
+    def fun() -> Iterator[int]:
+        yield from [1, 2, 3]
+
+    # this tests that it doesn't crash
+    # for actual version upgrade test see test_version_change
+    assert [1, 2, 3] == list(fun())
